/*$Id: modules.c,v 1.15 2002/12/20 10:59:40 aa5779 Exp $*/
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef USE_LTDL
#include <ltdl.h>
#endif
#include <string.h>
#include <ctype.h>
#include "dstring.h"
#include "util.h"
#include "lists.h"
#include "modules.h"
#include "interp.h"

static library *ext_libs;
static module *all_modules;
static denied_module *denied_list; 
static dstring *libpath;
extern char *nsl_location;

DEFLIST(mod_dependency,
		unsigned char is_lib : 1;
		unsigned char is_deny : 1;
		char *depname;
	);

DEFLIST(module_map,
		char *intname;
		char *extname;
		mod_dependency *deps;
		);
static module_map *map_of_modules;

static int mmap_find(module_map *m, char *name)
{
	return !strcmp(m->intname, name);
}

int module_in_map(char *name)
{
	return list_find(map_of_modules, name, (list_predicate_t)mmap_find) != NULL;
}

static int lib_find(library *lib, char *name)
{
	return !strcmp(lib->name, name);
}

library *load_library(char *name)
{
	library *lib = list_find(ext_libs, name, (list_predicate_t)lib_find);
	dstring *fname;
	if(lib)
		return lib;
#ifndef USE_LTDL
	recover_error("loadable module support is off");
#else	
	lib = ALLOC(library);
	fname = ds_create("lib");
	ds_appendstr(fname, name);
	lib->name = xstrdup(name);
	lib->handle = lt_dlopenext(DS_BODY(fname));
	if(!lib->handle)
		recover_error("can't load lib%s: %s", name, lt_dlerror());
	ext_libs = list_add(ext_libs, lib);
	return lib;
#endif	
}

static int mod_find(module *mod, char *name)
{
	return !strcmp(mod->name, name);
}

static interpreter *denied_in;
static int denied_find(denied_module *mod, char *name)
{
	return !strcmp(mod->name, name) && mod->i == denied_in;
}

module *load_module(char *name)
{
	module_map *mm;
	mod_dependency *dep;
	module *mod = list_find(all_modules, name, (list_predicate_t)mod_find);
	mod_init_f initf;
	
	if(mod)
	{
		if(!mod->handle && mod->map_ptr->extname)
			recover_error("module circular reference: %s", name);
		return mod;
	}
#ifndef USE_LTDL
	recover_error("loadable module support is off");
#else	
	denied_in = NULL;
	if(list_find(denied_list, name, (list_predicate_t)denied_find))
		recover_error("module '%s' is denied", name);
	mm = list_find(map_of_modules, name, (list_predicate_t)mmap_find);
	if(!mm)
		recover_error("module '%s' not installed", name);
	mod = ALLOC(module);
	mod->name = xstrdup(name);
	mod->map_ptr = mm;
	all_modules = list_add(all_modules, mod);
	for(dep = mm->deps; dep; dep = list_next(dep))
	{
		if(dep->is_deny)
			deny_module(NULL, dep->depname, deny_loaded | deny_loading);
		else if(dep->is_lib)
#ifdef RESOLVE_LIBDEPS		
			load_library(dep->depname);
#else
		/* do nothing */ ;
#endif
		else
			load_module(dep->depname);
	}
	if(mm->extname)
	{
		mod->handle = lt_dlopenext(mm->extname);
		if(!mod->handle)
			recover_error("can't load module: %s (%s): %s", name, mm->extname, lt_dlerror());
		initf = lt_dlsym(mod->handle, "init");
		if(initf)
			initf(mod);
	}
	return mod;
#endif	
}

static int att_find(mod_attach *ma, interpreter *i)
{
	return ma->i == i;
}

void attach_module(interpreter *i, char *name)
{	
	module *mod = load_module(name);
	mod_attach *ma;
	mod_attach_f attachf;
	if(list_find(mod->attached, i, (list_predicate_t)att_find))
		return;
	denied_in = i;
	if(list_find(denied_list, name, (list_predicate_t)denied_find))
		recover_error("'%s' is denied", name);
	ma = ALLOC(mod_attach);
	ma->i = i;
#ifdef USE_LTDL	
	if(mod->handle)
	{
		attachf = lt_dlsym(mod->handle, "attach");
		if(attachf)
			attachf(ma);
	}
	else
	{
		mod_dependency *dep;
		for(dep = mod->map_ptr->deps; dep; dep = list_next(dep))
		{
			if(!dep->is_lib && !dep->is_deny)
				attach_module(i, dep->depname);
		}
	}
#endif	
	mod->attached = list_add(mod->attached, ma);
}

void module_pragma(interpreter *i, char *name, char *pragma)
{
#ifdef USE_LTDL	
	module *mod = list_find(all_modules, name, (list_predicate_t)mod_find);
	mod_attach *ma;
	void (*pragmaf)(mod_attach *ma, char *pragma);
	if(!mod)
		recover_error("module '%s' not loaded", name);
	if(!(ma = list_find(mod->attached, i, (list_predicate_t)att_find)))
		recover_error("module '%s' not attached", name);
	pragmaf = lt_dlsym(mod->handle, "pragma");
	if(pragmaf)
		pragmaf(ma, pragma);
#endif	
}

static const char *module_string(char *name, char *strid)
{
#ifdef USE_LTDL
	module *mod = list_find(all_modules, name, (list_predicate_t)mod_find);
	const char *str;
	if(!mod)
		recover_error("no such module: %s", name);
	str = lt_dlsym(mod->handle, strid);
	return str ? str : "";
#else
	return "";
#endif		
}

const char *module_version(char *name)
{
	return module_string(name, "version");
}

const char *module_date(char *name)
{
	return module_string(name, "date");
}

const char *module_srcdate(char *name)
{
	return translate_cvs_date(module_string(name, "srcdate"));
}

void deny_module(interpreter *i, char *name, int flags)
{
	if(flags & deny_loaded)
	{
		module *found = list_find(all_modules, name, (list_predicate_t)mod_find);
		if(!i)
		{
			if(found)
				recover_error("'%s' is loaded", name);
		}
		else
		{
			if(found && list_find(found->attached, i, (list_predicate_t)att_find))
				recover_error("'%s' is denied", name);
		}
	}
	if(flags & deny_loading)
	{
		denied_module *dm  = ALLOC(denied_module);
		dm->name = xstrdup(name);
		dm->i = i;
		denied_list = list_add(denied_list, dm);
	}
}

static void finish_modules(void)
{
	module *miter;
	mod_attach *maiter;
	mod_attach_f detach;
	mod_init_f finish;
#ifdef USE_LTDL
	for(miter = all_modules; miter; miter = list_next(miter))
	{
		if(!miter->handle)
			continue;
		detach = lt_dlsym(miter->handle, "detach");
		if(detach)
		{
			for(maiter = miter->attached; maiter; maiter = list_next(maiter))
				detach(maiter);
		}
		finish = lt_dlsym(miter->handle, "finish");
		if(finish)
			finish(miter);
	}
	lt_dlexit();
#endif	
}

#ifdef USE_LTDL
static void process_module_map(void)
{
	char *file = load_file("module.map", nsl_location, "");
	char *iter = file;
	module_map *nm = NULL;
	mod_dependency *dep = NULL;
	dstring *tmp = ds_create(NULL);
	for(;;)
	{
		skip_spaces(iter, &iter);
		if(!*iter)
			break;
		nm = ALLOC(module_map);
		nm->intname = xstrdup(parse_id(iter, &iter));
		skip_spaces(iter, &iter);
		ds_assign(tmp, NULL);
		while(*iter && !isspace(*iter))
			ds_appendch(tmp, *iter++);
		if(!ds_comparestr(tmp, "idem", NULL, NULL) || !ds_comparestr(tmp, "=", NULL, NULL))
		{
			ds_assign(tmp, NULL);
			ds_appendstr(tmp, nm->intname);
		}
		nm->extname = !ds_comparestr(tmp, "wrapper", NULL, NULL) ? NULL : xstrdup(DS_BODY(ds_concat(libpath, tmp)));
		for(;;)
		{
			skip_spaces(iter, &iter);
			if(!*iter || *iter == ';')
			{
				iter += (*iter == ';');
				break;
			}
			dep = ALLOC(mod_dependency);
			if(*iter == '@' || *iter == '!')
			{
				dep->is_deny = *iter == '!';
				iter++;
				dep->depname = xstrdup(parse_id(iter, &iter));
			}
			else
			{
				if(*iter == '-' && iter[1] == 'l')
					iter += 2;
				dep->depname = xstrdup(parse_id(iter, &iter));
				dep->is_lib = 1;
			}
			nm->deps = list_add(nm->deps, dep);
		}
		map_of_modules = list_add(map_of_modules, nm);
	}
	free(file);
}
#endif

void init_modules(char *dflt)
{
#ifdef USE_LTDL	
	libpath = ds_create(getenv(SL_MOD_PATH_ENV));
	LTDL_SET_PRELOADED_SYMBOLS();
	lt_dlmalloc = xmalloc;
	if(lt_dlinit())
		fatal_error("can't initialize LTDL: %s", lt_dlerror());
	if(!ds_isempty(libpath))
		ds_fix(libpath);
	else
	{
		dstring *tmp = ds_fix(ds_create(nsl_location));
		ds_appendstr(tmp, "modules/");
		libpath = tmp;
	}
	process_module_map();
	load_module(dflt ? dflt : "preload");
	enh_atexit(finish_modules);
#endif	
}

void *get_mod_sym(char *modname, char *sym)
{
#ifdef USE_LTDL	
	module *mod = load_module(modname);
	return lt_dlsym(mod->handle, sym);
#else
	return NULL;
#endif	
}

void *get_ma_data(module *mod, interpreter *i)
{
	mod_attach *ma = list_find(mod->attached, i, (list_predicate_t)att_find);
	return ma ? ma->data : NULL;
}

void append_builtins(interpreter *i, master_instruction_src *set)
{
	master_instruction_src *iter;
	int count = 0;
	for(iter = set; iter->name; iter++, count++)
	{
		if(lookup_instruction(i, iter->name))
			recover_error("'%s' already defined", iter->name);
	}
	merge_builtins(i, set);
}

void replace_builtins(interpreter *i, master_instruction_src *rep, instruction_t *saved)
{
	instruction_src *dest;
	for(; rep->name; rep++, saved++)
	{
		dest = lookup_instruction(i, rep->name);
		if(!dest)
			recover_error("'%s' undefined", rep->name);
		*saved = dest->action;
		dest->action = rep->action;
	}
}

